"""
This module contains description of the function
that allows to initialize custom classes with clear
and informative `repr` method and base class with such method.

"""


from typing import Any, Iterable, Optional, Set


def remove_private_prefix(attr_name: str, prefix: str) -> str:
    """Return the specified attribute name without the specified prefix."""
    return attr_name[len(prefix):]


def get_representation(
        instance: Any,
        excluded: Optional[Set[str]] = None,
        ancestor_private_attributes: bool = True,
        is_only_tuples: bool = True,
) -> str:
    """
    Return the 'official' string representation of the specified instance.

    Parameters
    ----------
    instance : Any
        The instance, which string representation is returned.
    excluded : set, optional
        Names of arguments that are excluded
        from the representation.
    ancestor_private_attributes : bool, optional, default: True
        Private attributes of ancestor classes without '__' prefix
        are included into string representation, if this parameter is True,
        and are skipped if it is False.
    is_only_tuples : bool, optional, default: True
        Iterable attributes are represented as tuples,
        if this parameter is True.

    Returns
    -------
    str
        The 'official' string representation of the specified instance.

    """
    _class_name = instance.__class__.__name__
    representation = [f'{_class_name}(']
    for _key, _value in instance.__dict__.items():
        _key_repr, _value_repr = _key, _value
        _ancestor = instance.__class__.__bases__[0]
        while _ancestor.__name__ != 'object':
            _prefix = f'_{_ancestor.__name__}__'
            if _key.startswith(_prefix):
                if ancestor_private_attributes:
                    _key_repr = remove_private_prefix(_key, prefix=_prefix)
                else:
                    _key_repr = ''
            _ancestor = _ancestor.__bases__[0]
        if _key_repr == '':
            continue
        _prefix = f'_{_class_name}__'
        if _key.startswith(_prefix):
            _key_repr = remove_private_prefix(_key, prefix=_prefix)
        if excluded and _key_repr in excluded:
            continue
        if (
                is_only_tuples
                and isinstance(_value, Iterable)
                and not isinstance(_value, str)
        ):
            _value_repr = tuple(_value)
        representation.append(f'{_key_repr}={_value_repr!r}')
        representation.append(', ')
    if len(representation) > 1:
        representation.pop()
    representation.append(')')

    return ''.join(representation)


class RepresentableObject:
    """Class with custom representation."""

    def __repr__(self) -> str:
        """Return the 'official' string representation of instance."""
        return get_representation(
            self,
            ancestor_private_attributes=True,
            is_only_tuples=True,
            excluded=self.excluded_attributes_for_repr
        )

    def __str__(self) -> str:
        """Return the nicely printable string representation of instance."""
        return repr(self)

    @property
    def excluded_attributes_for_repr(self) -> Set[str]:
        """Return attributes that are not shown in instance representation."""
        return set()



"""
This package contains description of the function
that allows to initialize custom classes with clear
and informative `repr` method and base class with such method.

"""


from .main import get_representation, RepresentableObject



"""
This module is used for checking the project code for compliance with PEP8.

"""


from pylint_af import PyLinter


if __name__ == '__main__':
    PyLinter(
        is_printed=False,
        ignored_statements={
            'E0401', 'E0402',
            'C0115', 'C0116',
            'W0201', 'R0903',
        },
    ).check()



"""This module contain test for `pretty_repr` project."""


from .main import get_representation, RepresentableObject


class Parent(RepresentableObject):
    def __init__(self, attr_1, attr_2, default_1=1):
        """Initialize self. See help(type(self)) for accurate signature."""
        self.attr_1 = attr_1
        self.attr_2 = attr_2
        self.default_1 = default_1

    @property
    def attr_2(self):
        return self.__attr_2

    @attr_2.setter
    def attr_2(self, attr_2):
        self.__attr_2 = attr_2


class Child(Parent):
    def __init__(self, attr_3, attr_4, default_2=2, **kwargs):
        """Initialize self. See help(type(self)) for accurate signature."""
        super().__init__(**kwargs)
        self.attr_3 = attr_3
        self.attr_4 = attr_4
        self.default_2 = default_2

    @property
    def attr_4(self):
        return self.__attr_4

    @attr_4.setter
    def attr_4(self, attr_4):
        self.__attr_4 = attr_4


class GrandChild(Child):
    def __init__(self, attr_5, attr_6, default_3=3, **kwargs):
        """Initialize self. See help(type(self)) for accurate signature."""
        super().__init__(**kwargs)
        self.attr_5 = attr_5
        self.attr_6 = attr_6
        self.default_3 = default_3

    @property
    def attr_6(self):
        return self.__attr_6

    @attr_6.setter
    def attr_6(self, attr_6):
        self.__attr_6 = attr_6


class GrandChildExcluded(GrandChild):

    @property
    def excluded_attributes_for_repr(self):
        """Return attributes that are not shown in instance representation."""
        return {'attr_1', 'attr_4', 'attr_6'}


class ClassWithIterableAttributes(RepresentableObject):

    def __init__(self, attr_1, attr_2, attr_3, attr_4):
        """Initialize self. See help(type(self)) for accurate signature."""
        self.attr_1 = attr_1
        self.attr_2 = attr_2
        self.attr_3 = attr_3
        self.attr_4 = attr_4


class TestPrettyRepr:

    def setup(self):
        self.parent = Parent(attr_1=1, attr_2=2)
        self.child = Child(attr_1=1, attr_2=2, attr_3=3, attr_4=4, default_1=2)
        _kwargs = {
            'attr_1': 1,
            'attr_2': 2,
            'attr_3': 3,
            'attr_4': 4,
            'attr_5': 5,
            'attr_6': 6,
        }
        self.grand_child = GrandChild(**_kwargs)
        self.grand_child_excluded = GrandChildExcluded(**_kwargs)
        self.class_with_iterable_attributes = ClassWithIterableAttributes(
            attr_1=(1, 2, 3),
            attr_2=[4, 5, 6],
            attr_3={1: 2, 3: 4, 5: 6}.items(),
            attr_4='hello',
        )

    def test_parent(self):
        _exp = 'Parent(attr_1=1, attr_2=2, default_1=1)'
        _got = get_representation(self.parent)
        assert _got == _exp

    def test_child(self):
        _exp = (
            'Child(attr_1=1, attr_2=2, default_1=2, '
            'attr_3=3, attr_4=4, default_2=2)'
        )
        _got = get_representation(self.child)
        assert _got == _exp

    def test_grand_child(self):
        _exp = (
            'GrandChild(attr_1=1, attr_2=2, default_1=1, '
            'attr_3=3, attr_4=4, default_2=2, attr_5=5, attr_6=6, default_3=3)'
        )
        _got = get_representation(self.grand_child)
        assert _got == _exp

    def test_grand_child_with_excluded(self):
        _exp = (
            'GrandChild(attr_2=2, default_1=1, '
            'attr_3=3, default_2=2, attr_5=5, default_3=3)'
        )
        _got = get_representation(
            self.grand_child,
            excluded={'attr_1', 'attr_4', 'attr_6'}
        )
        assert _got == _exp

    def test_parent_without_base_arguments(self):
        _exp = 'Parent(attr_1=1, attr_2=2, default_1=1)'
        _got = get_representation(
            self.parent,
            ancestor_private_attributes=False,
        )
        assert _got == _exp

    def test_child_without_base_arguments(self):
        _exp = 'Child(attr_1=1, default_1=2, attr_3=3, attr_4=4, default_2=2)'
        _got = get_representation(
            self.child,
            ancestor_private_attributes=False,
        )
        assert _got == _exp

    def test_grand_child_without_base_arguments(self):
        _exp = (
            'GrandChild(attr_1=1, default_1=1, attr_3=3, '
            'default_2=2, attr_5=5, attr_6=6, default_3=3)'
        )
        _got = get_representation(
            self.grand_child,
            ancestor_private_attributes=False,
        )
        assert _got == _exp

    def test_parent_repr(self):
        _exp = 'Parent(attr_1=1, attr_2=2, default_1=1)'
        assert repr(self.parent) == _exp

    def test_child_repr(self):
        _exp = (
            'Child(attr_1=1, attr_2=2, default_1=2, '
            'attr_3=3, attr_4=4, default_2=2)'
        )
        assert repr(self.child) == _exp

    def test_grand_child_repr(self):
        _exp = (
            'GrandChild(attr_1=1, attr_2=2, default_1=1, '
            'attr_3=3, attr_4=4, default_2=2, attr_5=5, attr_6=6, default_3=3)'
        )
        assert repr(self.grand_child) == _exp

    def test_grand_child_repr_with_excluded(self):
        _exp = (
            'GrandChildExcluded(attr_2=2, default_1=1, '
            'attr_3=3, default_2=2, attr_5=5, default_3=3)'
        )
        _got = repr(self.grand_child_excluded)
        assert _got == _exp

    def test_class_with_iterable_attributes(self):
        _exp = (
            'ClassWithIterableAttributes(attr_1=(1, 2, 3), attr_2=[4, 5, 6], '
            f"attr_3=dict_items([(1, 2), (3, 4), (5, 6)]), attr_4='hello')"
        )
        _got = get_representation(
            self.class_with_iterable_attributes,
            is_only_tuples=False,
        )
        assert _got == _exp

    def test_class_with_iterable_attributes_as_tuples(self):
        _exp = (
            'ClassWithIterableAttributes(attr_1=(1, 2, 3), attr_2=(4, 5, 6), '
            f"attr_3=((1, 2), (3, 4), (5, 6)), attr_4='hello')"
        )
        _got = repr(self.class_with_iterable_attributes)
        assert _got == _exp



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

requirements = ["pylint-af==1.0.1", "pytest==6.2.5"]

setuptools.setup(
    name="pretty_repr",
    version="1.0.2",
    author="Albert Farkhutdinov",
    author_email="albertfarhutdinov@gmail.com",
    description=(
        "The package that allows you to create clear and informative methods "
        "for your own classes in Python."
    ),
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/AlbertFarkhutdinov/pretty_repr",
    packages=setuptools.find_packages(),
    install_requires=requirements,
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Programming Language :: Python :: 3.9",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)
